package com.keyware.composeanalysis.task;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.lang.Pair;
import com.keyware.common.constant.enums.AnalysisStatusEnum;
import com.keyware.composeanalysis.constant.FixedValue;
import com.keyware.composeanalysis.constant.RedisConst;
import com.keyware.composeanalysis.constant.enums.AnalysisLevelEnum;
import com.keyware.composeanalysis.constant.enums.FileAnalysisStatusEnum;
import com.keyware.composeanalysis.entity.AnalysisTask;
import com.keyware.composeanalysis.mongo.FileDataMongoDto;
import com.keyware.composeanalysis.mongo.MatchOpenFile;
import com.keyware.composeanalysis.mongo.MatchOpenFileMongoDto;
import com.keyware.composeanalysis.solr.VersionTree;
import com.keyware.composeanalysis.util.*;
import com.keyware.keyswan.anaysis.Analysis;
import com.keyware.keyswan.anaysis.AnalysisFactory;
import com.keyware.keyswan.common.CodeFile;
import com.keyware.utils.IdGenerator;
import lombok.extern.log4j.Log4j2;
import org.apache.solr.common.SolrDocument;
import org.apache.solr.common.SolrDocumentList;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Update;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

import static org.springframework.data.mongodb.core.query.Criteria.where;

/**
 * @author liuzongren
 * @date 2024/7/23
 * desc 文件级溯源分析任务
 */
@Log4j2
public class FileAnalysisTask extends IAnalysisTask {

    private MongoTemplate mongoTemplate;
    private AnalysisTask analysisTask;
    private SolrUtils solrUtils;
    //文件信息
    private FileDataMongoDto analysisFile;
    private RedisUtil redisUtil;
    private CountDownLatch countDownLatch;


    public FileAnalysisTask(AnalysisTask analysisTask, FileDataMongoDto analysisFile, MongoTemplate mongoTemplate, CountDownLatch countDownLatch) {
        this.mongoTemplate = mongoTemplate;
        this.analysisTask = analysisTask;
        this.analysisFile = analysisFile;
        this.countDownLatch = countDownLatch;
        this.solrUtils = SpringContextUtils.getBean(SolrUtils.class);
        this.redisUtil = SpringContextUtils.getBean(RedisUtil.class);
    }


    /**
     * 文件级溯源分析
     * 当前级别溯源分析 需要在 项目级级分析完成后执行
     * 当前文件源MD5 已经在solr库中匹配不到了，需要提取特征去匹配
     */
    @Override
    public void run() {
        //执行任务前,判断一下任务执行的状态
        Object status = redisUtil.get(String.format(RedisConst.TASK_RUNNING_STATUS_KEY_PREFIX, analysisTask.getId()));
        if (status != null && (status.equals(AnalysisStatusEnum.STOP_ANALYSIS.getCode()) || status.equals(AnalysisStatusEnum.PAUSE_ANALYSIS.getCode()))) {
            log.info("任务已取消,fileName:{}", analysisFile.getName());
            countDownLatch.countDown();
            return;
        }
        //获取当前文件名称
        String fileName = analysisFile.getName();

        try {
            //根据文件后缀 查询 *_CutFileInfo库名称
            String featureCoreName = FixedValue.SUFFIX_SOLR_FILE.get(analysisFile.getSuffix());
            //根据文件名称，获取文件解析器
            Analysis analysis = AnalysisFactory.getAnalysis(fileName);
            //如果文件大小超过3M,则不进行文件级行级特征提取
            CodeFile codeFile = analysis.analysisFile(analysisFile.getFileUrl(), "1", "0");

            //根据文件的特征值,去相应文件文件后缀的特征库中进行查询
            SolrDocumentList openSourceFileList = getFeatureSimilarityFromSolr(featureCoreName, codeFile);

            //则统计当前文件的开源率
            ananlyzeFileOpenRate(openSourceFileList, codeFile);

            //更新文件级分析结果
            analysisFile.setFileAnalysisStatus(FileAnalysisStatusEnum.ANALYSIS_DONE.getCode());
            mongoTemplate.update(FileDataMongoDto.class)
                    .matching(where("_id").is(analysisFile.getId()))
                    .replaceWith(analysisFile)
                    .findAndReplace();

            AnalysisLogUtil.insert(mongoTemplate, "【文件级分析】成功" + fileName);
        } catch (Exception e) {
            AnalysisLogUtil.insertErrorInfo(mongoTemplate, "【文件级】分析失败" + fileName, e);
            log.error("文件:" + fileName + "文件级别分析失败!", e);
            //将当前文件的分析状态变更为失败
            analysisFile.setFileAnalysisStatus(FileAnalysisStatusEnum.FAILED_ANALYSIS.getCode());
            //更新文件级分析结果
            mongoTemplate.update(FileDataMongoDto.class)
                    .matching(where("_id").is(analysisFile.getId()))
                    .apply(new Update().set("fileAnalysisStatus", FileAnalysisStatusEnum.FAILED_ANALYSIS.getCode()))
                    .first();
        } finally {
            countDownLatch.countDown();
        }
    }


    /**
     * 根据 特征值 从特征库中检索 具有特征相似的
     *
     * @param featureCoreName 检索的solr 库名称
     * @param codeFile        源文件解析结果
     * @return
     */
    private SolrDocumentList getFeatureSimilarityFromSolr(String featureCoreName, CodeFile codeFile) {
        if (codeFile == null) {
            log.error("特征为空,无法查询:{}", analysisFile.getName());
            return new SolrDocumentList();
        }
        String queryStr = "sourceMd5:" + codeFile.getSourceMd5() + " OR cutFileMd5:" + codeFile.getCutFileMd5() + " OR traitFileMd5:" + codeFile.getTraitFileMd5();
        return solrUtils.query(featureCoreName, queryStr, "sourceMd5");
    }

    /**
     * 分析文件的开源率
     *
     * @param openSourceFileList 匹配的开源文件信息
     * @throws IOException
     */
    private void ananlyzeFileOpenRate(SolrDocumentList openSourceFileList, CodeFile fileAnalysisRes) {

        if (CollUtil.isEmpty(openSourceFileList)){
            return;
        }

        HashSet<Integer> openLineNum = new HashSet<>();

        //计算每个文件的开源率和特征相似度
        List<MatchOpenFile> matchOpenFilesRes = calculateOpenRate(openSourceFileList, fileAnalysisRes, openLineNum);

        //获取开源率阈值，判断当前文件是否开源
        Integer openRateThreshold = analysisTask.getOpenRateThreshold();

        BigDecimal openRate = new BigDecimal(openLineNum.size()).divide(new BigDecimal(analysisFile.getCodeRowNum()), 4, RoundingMode.HALF_UP).multiply(new BigDecimal(100)).setScale(2);

        //超过阈值，则认为当前文件是开源文件
        if (openRate.compareTo(new BigDecimal(openRateThreshold)) > 0) {
            analysisFile.setOpenType(true);
        } else {
            analysisFile.setOpenType(false);
        }

        //修改保存测试文件信息
        analysisFile.setOpenLineCount(openLineNum.size())
                .setOpenRate(openRate.floatValue());


        //保存当前文件的开源信息到mongo库中
        MatchOpenFileMongoDto matchOpenFileInfo = new MatchOpenFileMongoDto();
        matchOpenFileInfo.setId(IdGenerator.uuid32())
                .setFileName(analysisFile.getName())
                .setFilePath(analysisFile.getFileUrl())
                .setOpenType(true)
                .setFeatureSimilarity(100.00f)
                .setOpenRate(openRate.floatValue())
                .setAnalysisType(AnalysisLevelEnum.FILE_LEVEL.getCode())
                .setSubMatchOpenFiles(matchOpenFilesRes);

        //保存当前开源信息数据
        mongoTemplate.insert(matchOpenFileInfo);

    }

    /**
     * 计算当前文件的特征相似度 和 开源率
     *
     * @param matchOpenFiles         通过MD5 匹配到的所有开源文件
     * @param matchLineRowsNum       所有开源文件匹配到的开源行号列表
     * @return 匹配的开源文件解析后的结果集
     */
    private List<MatchOpenFile> calculateOpenRate(SolrDocumentList matchOpenFiles, CodeFile fileAnalysisRes, Set<Integer> matchLineRowsNum) {

        //匹配的开源文件列表
        List<MatchOpenFile> matchOpenFilesRes = new ArrayList<>();

        //根据匹配的开源文件的md5 获取版本ID
        String sourceFileBaseCoreName = FixedValue.SUFFIX_SOLR_VERSION.get(analysisFile.getSuffix());

        //首先根据文件的MD5查询开源文件的版本ID,和路径信息
        Set<String> openSourceFileMd5s = matchOpenFiles.stream().map(doc -> (String) doc.get("sourceMd5")).collect(Collectors.toSet());
        Map<String, SolrDocument> md5VersionInfoMap = solrUtils.batchQueryVersionIdFromSourceFileBaseBySourceMd5(sourceFileBaseCoreName, openSourceFileMd5s);

        //根据版本ID查询版本的详细信息
        //todo 这里 查询一个版本的信息 需要检索 两个 solr 库 而且还需要检索 versioinTree 后面需要优化
        Set<String> openSourceFileVersionIds = md5VersionInfoMap.values().stream().map(doc -> (String) doc.get("versionId")).collect(Collectors.toSet());
        List<VersionTree> versionTrees = solrUtils.queryBatchVersionInfoByVersionIds(openSourceFileVersionIds);
        Map<String, VersionTree> versionIdVersionInfoMap = versionTrees.stream().collect(Collectors.toMap(VersionTree::getVersionId, java.util.function.Function.identity()));


        //被测件文本内容
        String sourcefileContent= FileUtil.readUtf8String(analysisFile.getFileUrl());

        //将文本内容解析成行信息,用于后续文件的开源率计算
        List<String> analysisFileLineInfo = SimilarityUtil.getSplitWords(sourcefileContent);

        for (SolrDocument openSourceFile : matchOpenFiles) {

            //开源文件md5
            String openSourceFileMd5 = openSourceFile.getFieldValue("sourceMd5").toString();

            String openFileContent = solrUtils.getOpenFileContentByMd5(openSourceFileMd5);

            //当前文件的开源率
            Pair<Float, HashSet<Integer>> openRateAndSaveRowNum = SimilarityUtil.getOpenRateAndSaveRowNum(analysisFileLineInfo, openFileContent);
            //将当前文件匹配的行号，存储到缓存中，方便统计整体的开源率
            matchLineRowsNum.addAll(openRateAndSaveRowNum.getValue());

            SolrDocument openEntries = md5VersionInfoMap.get(openSourceFileMd5);
            VersionTree versionInfo = versionIdVersionInfoMap.get(openEntries.get("versionId"));
            if (versionInfo == null) {
                log.error("找不到开源文件版本信息,versionId:{}", openEntries.get("versionId"));
            }

            String openFilePath = (String) openEntries.get("fullPath");
            //组装当前开源文件的开源项目信息
            MatchOpenFile matchOpenFile = new MatchOpenFile();
            matchOpenFile.setId(IdGenerator.uuid32())
                    .setFileName(FileUtil.getName(openFilePath))
                    .setPName(versionInfo.getProName())
                    .setPId(versionInfo.getProId())
                    .setVersion(versionInfo.getVersionName())
                    .setVersionId(versionInfo.getVersionId())
                    .setSourceFilePath(openFilePath)
                    .setSourceUrl(versionInfo.getDownUrl())
                    .setLicenseType(versionInfo.getLicenseType())
                    .setFeatureSimilarity(100.00f)
                    .setOpenRate(openRateAndSaveRowNum.getKey());
            matchOpenFilesRes.add(matchOpenFile);
        }
        return matchOpenFilesRes;
    }


}
